#if !defined HAVE_BITARRAY_H__
#define      HAVE_BITARRAY_H__


#include "fxttypes.h"
#include "bitsperlong.h"
//#include "newop.h"

//#define NO_BITS_USE_ASM
//
#ifndef NO_BITS_USE_ASM  // for automatic testing
#include "bitasm.h"
#else
#undef BITS_USE_ASM
#endif


#define  DIVMOD(n, d, bm) \
ulong d = n / BITS_PER_LONG; \
ulong bm = 1UL << (n % BITS_PER_LONG);

#define  DIVMOD_TEST(n, d, bm) \
ulong d = n / BITS_PER_LONG; \
ulong bm = 1UL << (n % BITS_PER_LONG); \
ulong t = bm & f_[d];


class bitarray
//
// bitarray class mostly for use as
// memory saving array of boolean values
// valid index is 0...nb_-1 (as usual in C arrays)
//
{
public:
    ulong *f_;   // bit bucket
    ulong nb_;   // number of bits
    ulong nfw_;  // number of words where all bits are used, may be zero
    ulong mp_;   // mask for partially used word if there is one, else zero
    // (ones are at the positions of the unused bits)

public:
    bitarray(ulong nbits)
    {
        nb_ = ( 0==nbits ? 1 : nbits );  // avoid trouble for zero length

        // nw_: number of words (incl. partially used), nw>=1
        ulong nw_ = nb_ / BITS_PER_LONG;       // number of words

        // nbl_: number of bits used in last (partially used) word, 0 if mw==mfw
        ulong nbl_ = nb_ - nw_*BITS_PER_LONG;  // number of bits used in last word
        nfw_ = nw_;  // number of fully used words

        if ( 0!=nbl_ )  // there is a partially used word
        {
            ++nw_;  // increase total number of words
            mp_ = (~0UL) >> (BITS_PER_LONG-nbl_);  // correct mask for last word
        }
        else  mp_ = 0UL;

        f_ = new ulong[nw_];  // finally allocate
    }


    ~bitarray()  { delete [] f_; }


    ulong test(ulong n)  const
    // test whether n-th bit set
    {
        if ( n>=nb_ )  return 0;
#if defined  BITS_USE_ASM
        return asm_bt(f_, n);
#else
        DIVMOD_TEST(n, d, bm);
        return  t;
#endif
    }


    void set(ulong n)
    // set n-th bit
    {
        if ( n>=nb_ )  return;
#if defined  BITS_USE_ASM
        asm_b_s(f_, n);
#else
        DIVMOD(n, d, bm);
        f_[d] |= bm;
#endif
    }


    void clear(ulong n)
    // clear n-th bit
    {
        if ( n>=nb_ )  return;
#if defined  BITS_USE_ASM
        asm_b_r(f_, n);
#else
        DIVMOD(n, d, bm);
        f_[d] &= ~bm;
#endif
    }

    void change(ulong n)
    // toggle n-th bit
    {
        if ( n>=nb_ )  return;
#if defined  BITS_USE_ASM
        asm_b_c(f_, n);
#else
        DIVMOD(n, d, bm);
        f_[d] ^= bm;
#endif
    }


    ulong test_set(ulong n)
    // test whether n-th bit is set and set it
    {
        if ( n>=nb_ )  return 0;
#if defined  BITS_USE_ASM
        return asm_bts(f_, n);
#else
        DIVMOD_TEST(n, d, bm);
        f_[d] |= bm;
        return  t;
#endif
    }


    ulong test_clear(ulong n)
    // test whether n-th bit is set and clear it
    {
        if ( n>=nb_ )  return 0;
#if defined  BITS_USE_ASM
        return asm_btr(f_, n);
#else
        DIVMOD_TEST(n, d, bm);
        f_[d] &= ~bm;
        return  t;
#endif
    }


    ulong test_change(ulong n)
    // test whether n-th bit is set and toggle it
    {
        if ( n>=nb_ )  return 0;
#if defined  BITS_USE_ASM
        return asm_btc(f_, n);
#else
        DIVMOD_TEST(n, d, bm);
        f_[d] ^= bm;
        return  t;
#endif
    }


    void clear_all()
    // clear all bits
    {
        for (ulong k=0; k<nfw_; ++k)  f_[k] = 0;
        if ( mp_ )  f_[nfw_] = 0;
    }

    void set_all()
    // set all bits
    {
        for (ulong k=0; k<nfw_; ++k)  f_[k] = ~0UL;
        if ( mp_ )  f_[nfw_] = ~0UL;
    }


    ulong all_set_q()  const
    // return whether all bits are set
    {
        for (ulong k=0; k<nfw_; ++k)  if ( ~f_[k] )  return 0;
        if ( mp_ )
        {
            ulong z = f_[nfw_] & mp_;
            if ( z!=mp_ )  return 0;
        }
        return  1;
    }

    ulong all_clear_q()  const
    // return whether all bits are clear
    {
        for (ulong k=0; k<nfw_; ++k)  if ( f_[k] )  return 0;
        if ( mp_ )
        {
            ulong z = f_[nfw_] & mp_;
            if ( z!=0 )  return 0;
        }
        return  1;
    }

    ulong next_set_idx(ulong n)  const
    // return index of next set or value beyond end
    // Note: the given index n is included in the search
    {
        if ( n>=nb_ )  n = nb_;
        else while ( (n<nb_) && (!test(n)) )  ++n;
        return  n;
    }

    ulong next_clear_idx(ulong n)  const
    // return index of next clear or value beyond end
    // Note: the given index n is included in the search
    {
        if ( n>=nb_ )  n = nb_;
        else while ( (n<nb_) && (test(n)) )  ++n;
        return  n;
    }

    // todo: next_set_idx_sparse() / next_clear_idx_sparse()


    void dump()  const;
    void dump_bits()  const;
};
// -------------------------

#undef  DIVMOD
#undef  DIVMOD_TEST

//inline bitarray & operator ^= (bitarray &t, const bitarray &h)
//// no check for compatible lengths (!)
//{
//    ulong nfw = t.nfw_;
//    for (ulong k=0; k<nfw; ++k)  t.f_[k] ^= h.f_[k];
//    if ( t.mp_ )  t.f_[nfw] ^= h.f_[nfw];
//    return  t;
//}
//// -------------------------

#endif // !defined HAVE_BITARRAY_H__
