#if !defined  HAVE_BITINVOLUTIONS_H__
#define       HAVE_BITINVOLUTIONS_H__

#include "fxttypes.h"
#include "revbin.h"


inline ulong blue_code(ulong x)
// self-inverse
// range [0..2^k-1] is mapped onto itself
// work \prop log_2(BITS_PER_LONG)
//.
// With  B:=blue,  G:=gray,  I:=inverse_gray
//   G(B(G(B(x)))) == x
//   G(B(G(x)))    == B(x)
//   I(B(I(x)))    == B(x)
//   G(B(x))       == B(I(x))
// With  P:=parity
//   P(B(x)) == x % 2
{
    ulong s = BITS_PER_LONG >> 1;
    ulong m = ~0UL << s;
    while ( s )
    {
        x ^= ( (x&m) >> s );
        s >>= 1;
        m ^= (m>>s);
    }
    return  x;
}
// -------------------------


inline ulong yellow_code(ulong x)
// self-inverse
// work \prop log_2(BITS_PER_LONG)
//.
// With  Y:=yellow,  B:=blue,  r:=revbin
//   B(x) == Y(r(Y(x)))
//   Y(x) == r(B(r(x)))
//   r(x) == Y(B(Y(x))) == B(Y(B(x)))
// With  P:=parity
//   P(Y(x)) == highest_bit(x) == "sign(x)"
{
    ulong s = BITS_PER_LONG >> 1;
    ulong m = ~0UL >> s;
    while ( s )
    {
        x ^= ( (x&m) << s );
        s >>= 1;
        m ^= (m<<s);
    }
    return  x;
}
// -------------------------


inline ulong red_code(ulong x)
// self-inverse
// range [0..2^k-1] is mapped onto itself
//.
// With  R:=red,  B:=blue,  G:=gray,  I:=inverse_gray
//   G(R(x)) == I(B(x))
//     R(x)  == I(I(B(x)))
// todo:  version \prop log_2(BITS_PER_LONG)
{
    ulong m = ~0UL;
    ulong r = x & 1;
    ulong h = 2;
    ulong s = 2;
    do
    {
        if ( h & x )  r ^= (m >> (BITS_PER_LONG-s));
        h <<= 1;
        m ^= (m>>1);
        ++s;
    }
    while ( h );
    return  r;
}
// -------------------------


inline ulong cyan_code(ulong x)
// self-inverse
//.
// The relation between  red <--> cyan
//   is equivalent to    blue <--> yellow
// todo:  version \prop log_2(BITS_PER_LONG)
{
    return  red_code( revbin(x) );
}
// -------------------------


#endif  // !defined HAVE_BITINVOLUTIONS_H__
