#if !defined  HAVE_HEAPSORT_H__
#define       HAVE_HEAPSORT_H__

#include "heap.h"

#include "inline.h"
#define swap fxtaux::swap // avoid unwanted std::swap


template <typename Type>
void heap_sort_ascending(Type *x, ulong n)
// sort an array that has the heap-property into ascending order.
// On return x[] is _not_ a heap anymore.
{
    Type *p = x - 1;
    for (ulong k=n; k>1; --k)
    {
        swap(p[1], p[k]);   // move largest to end of array
        --n;                // remaining array is one element less
        heapify1(p, n, 1);  // restore heap-property
    }
}
// -------------------------

template <typename Type>
void heap_sort_decending(Type *x, ulong n)
// sort an array that has the heap-property into descending order.
// On return x[] is _not_ a heap anymore.
{
    Type *p = x - 1;
    for (ulong k=n; k>1; --k)
    {
        ++p;  --n;          // remaining array is one element less
        heapify1(p, n, 1);  // restore heap-property
    }
}
// -------------------------

template <typename Type>
void heap_sort(Type *x, ulong n)
{
    build_heap(x, n);
    heap_sort_ascending(x, n);
}
// -------------------------


#undef swap // end (avoid unwanted std::swap)

#endif  // !defined HAVE_HEAPSORT_H__
